#include "BLEDevice.h"

static BLEUUID     advertisedUUID("22bb746f-2ba0-7554-2d6f-726568705327");     // UUID advertised by the device
static BLEUUID        serviceUUID("22bb746f-2ba0-7554-2d6f-726568705327");     // UUID of the service containing a read characteristic. This will often be the same at the advertisedUUID, but not always.
static BLEUUID characteristicUUID("22bb746f-2ba1-7554-2d6f-726568705327");     // UUID of the caracteristic that can be read

int LED_PIN = 2;              // Set this to the pin number of your LED. The Firebeetle ESP32 onboard LED pin is 2
int keepAliveInterval = 0;    // Adjust this if your device requires a heartbeat to keep the connection alive. This defines the value in ms between calls. 0 means no heartbeat required.

static boolean doConnect = false;
static boolean connected = false;
static boolean doScan = false;
unsigned long blinkTimer = 0;
unsigned long lastblink;
unsigned long lastKeepAlive = 0;
int ledFlag = 0;
int rssi;
static BLERemoteCharacteristic* pRemoteCharacteristic;
static BLEAdvertisedDevice* myDevice;
BLEScan* pBLEScan;
BLEClient*  pClient;

// -----

class MyClientCallback : public BLEClientCallbacks {
  void onConnect(BLEClient* pclient) {
    Serial.println(" - Connected to server");
  }

  void onDisconnect(BLEClient* pclient) {
    connected = false;
    Serial.println("onDisconnect");
    delay(1000);
    Serial.println("RESETTING");
    ESP.restart();
  }
};

bool connectToServer() {
    Serial.print("Forming a connection to ");
    Serial.println(myDevice->getAddress().toString().c_str());
    
    pClient  = BLEDevice::createClient();
    Serial.println(" - Created client");

    pClient->setClientCallbacks(new MyClientCallback());

    // Connect to the remote BLE Server.
    Serial.println(pClient->connect(myDevice));
    

    // Obtain a reference to the service we are after in the remote BLE server.
    BLERemoteService* pRemoteService = pClient->getService(serviceUUID);
    if (pRemoteService == nullptr) {
      Serial.print("Failed to find our service UUID: ");
      Serial.println(advertisedUUID.toString().c_str());
      pClient->disconnect();
      return false;
    }
    Serial.println(" - Found our service");

    // Obtain a reference to the characteristic in the service of the remote BLE server.
    pRemoteCharacteristic = pRemoteService->getCharacteristic(characteristicUUID);
    if (pRemoteCharacteristic == nullptr) {
      Serial.print("Failed to find our characteristic UUID: ");
      Serial.println(characteristicUUID.toString().c_str());
      pClient->disconnect();
      return false;
    }
    Serial.println(" - Found our characteristic");

    connected = true;
    return true;
}

class MyAdvertisedDeviceCallbacks: public BLEAdvertisedDeviceCallbacks {
  void onResult(BLEAdvertisedDevice advertisedDevice) {
    Serial.print("BLE Advertised Device found: ");
    Serial.println(advertisedDevice.toString().c_str());

    // We have found a device, let us now see if it contains the service we are looking for.
    if (advertisedDevice.haveServiceUUID() && advertisedDevice.isAdvertisingService(advertisedUUID)) {

      pBLEScan->stop();
      myDevice = new BLEAdvertisedDevice(advertisedDevice);
      doConnect = true;
      doScan = true;

    }
  }
};

void handleBlink() {
  if (lastblink + blinkTimer <= millis() && ledFlag == 0 && rssi != 0) {
    lastblink = millis();
    digitalWrite(LED_PIN, HIGH);
    ledFlag = 1;
    
    Serial.print("Setting HIGH -- RSSI: ");
    Serial.println(rssi);
    
  } else if (lastblink + blinkTimer <= millis() && ledFlag == 1 && rssi != 0) {
    lastblink = millis();
    digitalWrite(LED_PIN, LOW);
    ledFlag = 0;
    Serial.print("Setting LOW -- RSSI: ");
    Serial.println(rssi);
  }
}

void setup() {
  Serial.begin(115200);
  Serial.println("Starting Arduino BLE Client application...");
  BLEDevice::init("");

  // Retrieve a Scanner and set the callback we want to use to be informed when we
  // have detected a new device.  Specify that we want active scanning and start the
  // scan to run for 5 seconds.
  pBLEScan = BLEDevice::getScan();
  pBLEScan->setAdvertisedDeviceCallbacks(new MyAdvertisedDeviceCallbacks());
  pBLEScan->setInterval(1349);
  pBLEScan->setWindow(449);
  pBLEScan->setActiveScan(true);
  pBLEScan->start(10, false);
}

void loop() {
  pinMode(LED_PIN, OUTPUT);
  // If the flag "doConnect" is true then we have scanned for and found the desired
  // BLE Server with which we wish to connect.  Now we connect to it.  Once we are 
  // connected we set the connected flag to be true.
  if (doConnect == true) {
    if (connectToServer()) {
      Serial.println("We are now connected to the BLE Server.");
    } else {
      Serial.println("We have failed to connect to the server; there is nothin more we will do.");
    }
    doConnect = false;
  }

  // If we are connected to a peer BLE Server, get the RSSI data and handle the LED. Keep the connection alive if necessary 
  if (connected) {
    if (keepAliveInterval > 0 && lastKeepAlive + keepAliveInterval <= millis()) {
      String val = pRemoteCharacteristic->readValue().c_str();  // Keep connection alive. Some BLE tags tend to go into low power mode, affecting RSSI
      lastKeepAlive = millis();      
    }
    rssi = abs(pClient->getRssi());
    Serial.print("RSSI: ");
    Serial.println(rssi);
    if (rssi <= 50) {                               // Adjust these values to change the speed your light blinks
      blinkTimer = map(rssi, 30, 50, 0, 300);       // map(low RSSI, high RSSI, shortest delay between blinks, longest delay between blinks)
    } else if (rssi <= 50) {
      blinkTimer = map(rssi, 50, 60, 300, 500);      
    } else if (rssi <= 60) {
      blinkTimer = map(rssi, 60, 70, 500, 800);      
    } else if (rssi <= 70) {
      blinkTimer = map(rssi, 70, 90, 800, 2000);      
    } else if (rssi <= 90) {
      blinkTimer = map(rssi, 90, 1000, 2000, 5000);      
    }
    handleBlink();
  }else if(doScan){
    BLEDevice::getScan()->start(0);
  }
}
